# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Callable extensions."""

from typing import Any, Callable, List

from pyglove.core.typing import callable_signature


class CallableWithOptionalKeywordArgs:
  """Helper class for invoking callable objects with optional keyword args.

  Examples::

    f = pg.typing.CallableWithOptionalKeywordArgs(lambda x: x ** 2, 'y')
    # Returns 4. Keyword 'y' is ignored.
    f(2, y=3)
  """

  def __init__(self,
               func: Callable[..., Any],
               optional_keywords: List[str]):
    sig = callable_signature.signature(
        func, auto_typing=False, auto_doc=False
    )

    # Check for variable keyword arguments.
    if sig.has_varkw:
      absent_keywords = []
    else:
      all_keywords = set(sig.arg_names)
      absent_keywords = [k for k in optional_keywords if k not in all_keywords]
    self._absent_keywords = absent_keywords
    self._func = func

  def __call__(self, *args, **kwargs):
    """Delegate the call to function."""
    for k in self._absent_keywords:
      kwargs.pop(k, None)
    return self._func(*args, **kwargs)



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.geno.Float."""

import random
import unittest

from pyglove.core.geno.base import DNA
from pyglove.core.geno.categorical import oneof
from pyglove.core.geno.numerical import floatv
from pyglove.core.geno.space import constant


class FloatTest(unittest.TestCase):
  """Tests for `pg.geno.Float`."""

  def test_init(self):
    floatv(-1.0, 1.0)
    floatv(0.0, 1.0, scale='linear')
    floatv(1e-8, 1.0, scale='log')
    floatv(1e-8, 1.0, scale='rlog')

    with self.assertRaisesRegex(
        ValueError,
        'Argument \'min_value\' .* should be no greater than \'max_value\''):
      floatv(1.0, -1.0)

    with self.assertRaisesRegex(ValueError, '\'min_value\' must be positive'):
      floatv(min_value=0.0, max_value=1.0, scale='log')

    with self.assertRaisesRegex(
        ValueError, 'Value .* is not in candidate list'):
      floatv(min_value=0.0, max_value=1.0, scale='unsupported_scale')

  def test_basics(self):
    x = floatv(0.1, 1.0, location='a.b')
    self.assertEqual(x.location.keys, ['a', 'b'])
    self.assertIsNone(x.scale)
    self.assertIsNone(x.name)

    self.assertFalse(x.is_space)
    self.assertFalse(x.is_categorical)
    self.assertFalse(x.is_subchoice)
    self.assertTrue(x.is_numerical)
    self.assertFalse(x.is_custom_decision_point)

  def test_len(self):
    self.assertEqual(len(floatv(0.1, 1.0)), 1)

  def test_space_size(self):
    self.assertEqual(floatv(0.1, 1.0).space_size, -1)

  def test_id(self):
    x = floatv(-1.0, 1.0, location='x')
    a = oneof([constant(), x], location='a')
    self.assertEqual(a.id, 'a')
    self.assertEqual(x.id, 'a[=1/2].x')

  def test_get(self):
    x = floatv(-1.0, 1.0, location='x', name='foo')
    a = oneof([x, constant()], location='a')

    # Get by ID.
    self.assertIs(a.get('a'), a)
    self.assertIs(a.get('a[=0/2].x'), x)
    self.assertIs(x.get('a[=0/2].x'), x)

    # Get by name.
    self.assertIs(a.get('foo'), x)

  def test_getitem(self):
    x = floatv(-1.0, 1.0, location='x', name='foo')
    _ = oneof([x, constant()], location='a')

    # Get by ID.
    self.assertIs(x['a[=0/2].x'], x)

    # Get by name.
    self.assertIs(x['foo'], x)

    with self.assertRaises(KeyError):
      _ = x['y']    # There is no decision point named `y` under x.

  def test_decision_ids(self):
    x = floatv(-1.0, 1.0, location='x')
    _ = oneof([x, constant()], location='a')
    self.assertEqual(
        x.decision_ids,
        [
            'a[=0/2].x',
        ])

  def test_decision_points(self):
    x = floatv(-1.0, 1.0, location='x')
    _ = oneof([x, constant()], location='a')
    self.assertEqual(x.decision_points, [x])

  def test_named_decision_points(self):
    x = floatv(-1.0, 1.0, location='x', name='foo')
    _ = oneof([x, constant()], location='a')
    self.assertEqual(x.named_decision_points, {'foo': x})

  def test_set_get_userdata(self):
    x = floatv(-1.0, 1.0)
    x.set_userdata('mydata', 1)
    self.assertEqual(x.userdata.mydata, 1)
    self.assertEqual(x.userdata['mydata'], 1)
    with self.assertRaises(KeyError):
      _ = x.userdata.nonexist_data
    with self.assertRaises(KeyError):
      _ = x.userdata['nonexist_data']

  def test_parent_spec(self):
    x = floatv(-1.0, 1.0, location='x')
    self.assertIsNone(x.parent_spec)
    a = oneof([x, constant()], location='a')
    self.assertIs(x.parent_spec, a.candidates[0])

  def test_parent_choice(self):
    x = floatv(-1.0, 1.0, location='x')
    self.assertIsNone(x.parent_choice)
    a = oneof([x, constant()], location='a')
    self.assertIs(x.parent_choice, a)

  def test_validate(self):
    x = floatv(min_value=0.0, max_value=1.0)
    x.validate(DNA(0.5))

    with self.assertRaisesRegex(ValueError, 'Expect float value'):
      x.validate(DNA(None))

    with self.assertRaisesRegex(ValueError, 'Expect float value'):
      x.validate(DNA(1))

    with self.assertRaisesRegex(
        ValueError, 'DNA value should be no less than 0.*'):
      x.validate(DNA(-1.0))

    with self.assertRaisesRegex(
        ValueError, 'DNA value should be no greater than 1.*'):
      x.validate(DNA(1.5))

    with self.assertRaisesRegex(
        ValueError, 'Float DNA should have no children'):
      x.validate(DNA((1.0, [1, 2])))

  def test_first_dna(self):
    x = floatv(min_value=0.0, max_value=1.0)
    self.assertEqual(x.first_dna(), DNA(0.0))

  def test_next_dna(self):
    x = floatv(min_value=0.0, max_value=1.0)
    self.assertEqual(x.next_dna(), DNA(0.0))
    with self.assertRaisesRegex(
        NotImplementedError, '`next_dna` is not supported on `Float` yet'):
      x.next_dna(DNA(0.0))

  def test_iter_dna(self):
    x = floatv(min_value=0.0, max_value=1.0)
    with self.assertRaisesRegex(
        NotImplementedError, '`next_dna` is not supported on `Float` yet'):
      list(x.iter_dna())

  def test_random_dna(self):
    x = floatv(min_value=0.0, max_value=1.0)
    self.assertEqual(
        x.random_dna(random.Random(1)), DNA(0.13436424411240122))

  def test_inspection(self):
    self.assertEqual(
        str(floatv(min_value=0.0, max_value=1.0, location='a.b')),
        'Float(id=\'a.b\', min_value=0.0, max_value=1.0)')


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.object_utils.formatting."""
import inspect
import unittest

from pyglove.core.object_utils import formatting


class Foo(formatting.Formattable):

  def format(
      self, compact: bool = False, verbose: bool = True, **kwargs):
    return f'{self.__class__.__name__}(compact={compact}, verbose={verbose})'


class Bar(formatting.Formattable):

  def __init__(self, foo: Foo):
    self._foo = foo

  def format(
      self, compact: bool = False, verbose: bool = True,
      root_indent: int = 0, **kwargs):
    foo_str = self._foo.format(
        compact=compact, verbose=verbose, root_indent=root_indent + 1)
    return f'{self.__class__.__name__}(foo={foo_str})'


class FormattableTest(unittest.TestCase):

  def test_formattable(self):
    foo = Foo()
    self.assertEqual(repr(foo), 'Foo(compact=True, verbose=True)')
    self.assertEqual(str(foo), 'Foo(compact=False, verbose=True)')

  def test_formattable_with_custom_format(self):
    class Baz(Foo):
      __str_format_kwargs__ = {'compact': False, 'verbose': False}
      __repr_format_kwargs__ = {'compact': True, 'verbose': False}

    bar = Baz()
    self.assertEqual(repr(bar), 'Baz(compact=True, verbose=False)')
    self.assertEqual(str(bar), 'Baz(compact=False, verbose=False)')

  def test_formattable_with_context_managers(self):
    foo = Foo()
    with formatting.str_format(verbose=False):
      with formatting.repr_format(compact=False):
        self.assertEqual(repr(foo), 'Foo(compact=False, verbose=True)')
        self.assertEqual(str(foo), 'Foo(compact=False, verbose=False)')


class StringHelperTest(unittest.TestCase):
  """Tests for string helper methods in formatting."""

  def test_raw_text(self):
    raw = formatting.RawText('abc')
    self.assertEqual(formatting.format(raw), 'abc')
    self.assertEqual(formatting.format(raw, compact=True), 'abc')
    self.assertEqual(raw, formatting.RawText('abc'))
    self.assertEqual(raw, 'abc')
    self.assertNotEqual(raw, formatting.RawText('abcd'))
    self.assertNotEqual(raw, 'abcd')

  def test_special_format_support(self):

    class NewLine:
      def _repr_html_(self):
        return '<hr>'

      def __str__(self):
        return 'NewLine()'

      def __repr__(self):
        return 'NewLine()'

    v = NewLine()
    self.assertEqual(formatting.str_ext(v), 'NewLine()')
    def _method(attr_name: str):
      def fn(v, root_indent):
        del root_indent
        f = getattr(v, attr_name, None)
        return f() if f is not None else None
      return fn

    self.assertEqual(
        formatting.str_ext(v, custom_format=_method('_repr_html_')),
        '<hr>'
    )
    self.assertEqual(
        formatting.str_ext(v, custom_format=_method('_repr_xml_')),
        'NewLine()'
    )
    self.assertEqual(formatting.repr_ext(v), 'NewLine()')
    self.assertEqual(
        formatting.repr_ext(v, custom_format=_method('_repr_html_')),
        '<hr>'
    )
    self.assertEqual(
        formatting.repr_ext(v, custom_format=_method('_repr_xml_')),
        'NewLine()'
    )

  def test_kvlist_str(self):
    self.assertEqual(
        formatting.kvlist_str([
            ('', 'foo', None),
            ('a', 1, None),
            ('b', 'str', (None, 'str')),
            ('c', [1, 2, 3], False),
        ], label='Foo'),
        'Foo(\'foo\', a=1, c=[1, 2, 3])'
    )

    self.assertEqual(
        formatting.kvlist_str([
            ('', 'foo', None),
            ('a', 1, None),
            ('b', 'str', (None, 'str')),
            ('c', True, False),
        ]),
        '\'foo\', a=1, c=True'
    )

    self.assertEqual(
        formatting.kvlist_str([
            ('', 'foo', None),
            ('a', 1, None),
            ('b', 'str', (None, 'str')),
            ('c', True, False),
        ], label='Foo', compact=False),
        'Foo(\n  \'foo\',\n  a=1,\n  c=True\n)'
    )

    self.assertEqual(
        formatting.kvlist_str([
            ('', 'foo', None),
            ('a', 1, None),
            ('b', 'str', (None, 'str')),
            ('c', dict(x=1), False),
        ], compact=False),
        '\'foo\',\na=1,\nc={\n  \'x\': 1\n}'
    )

    self.assertEqual(
        formatting.kvlist_str([
            ('', 'foo', 'foo')
        ], label='Foo', compact=False),
        'Foo()'
    )

    class Foo:    # pylint: disable=redefined-outer-name
      def _repr_xml_(self):
        return '<foo/>'

      def __str__(self):
        return 'Foo()'

    self.assertEqual(
        formatting.kvlist_str([
            ('', Foo(), None)
        ], compact=False),
        'Foo()'
    )
    def _custom_format(v, root_indent):
      del root_indent
      f = getattr(v, '_repr_xml_', None)
      return f() if f is not None else None

    self.assertEqual(
        formatting.kvlist_str([
            ('', Foo(), None)
        ], compact=False, custom_format=_custom_format),
        '<foo/>'
    )
    self.assertEqual(
        formatting.kvlist_str([
            ('', (Foo(), 1), None)
        ], compact=True, custom_format=_custom_format),
        '(<foo/>, 1)'
    )

  def test_quote_if_str(self):
    self.assertEqual(formatting.quote_if_str(1), 1)
    self.assertEqual(formatting.quote_if_str('foo'), '\'foo\'')
    self.assertEqual(formatting.quote_if_str('foo\'s\na'), '"foo\'s\\na"')

  def test_comma_delimited_str(self):
    self.assertEqual(
        formatting.comma_delimited_str([1, 2, 'abc']), '1, 2, \'abc\'')

  def test_auto_plural(self):
    self.assertEqual(formatting.auto_plural(2, 'number'), 'numbers')
    self.assertEqual(formatting.auto_plural(2, 'was', 'were'), 'were')


class FormatTest(unittest.TestCase):
  """Tests for formatting.format."""

  def test_formattable(self):

    class A(formatting.Formattable):

      def format(self, compact=True, **kwargs):
        if compact:
          return 'A()'
        else:
          return 'A(...)'

    self.assertEqual(str(A()), 'A(...)')
    self.assertEqual(repr(A()), 'A()')

  def test_simple_types(self):
    self.assertEqual(formatting.format(True, compact=True), 'True')
    self.assertEqual(formatting.format(1, compact=True), '1')
    self.assertEqual(formatting.format(1.0, compact=True), '1.0')
    self.assertEqual(formatting.format('foo', compact=True), '\'foo\'')
    self.assertEqual(
        formatting.format('foo\'s\na', compact=True), '"foo\'s\\na"')

    # Compact=False has no impact on simple types.
    self.assertEqual(formatting.format(True, compact=False), 'True')
    self.assertEqual(formatting.format(1, compact=False), '1')
    self.assertEqual(formatting.format(1.0, compact=False), '1.0')
    self.assertEqual(formatting.format('foo', compact=False), '\'foo\'')
    self.assertEqual(
        formatting.format('foo\'s\na', compact=False), '"foo\'s\\na"')

    # Verbose has no impact on simple types.
    self.assertEqual(formatting.format(True, verbose=True), 'True')
    self.assertEqual(formatting.format(1, verbose=True), '1')
    self.assertEqual(formatting.format(1.0, verbose=True), '1.0')
    self.assertEqual(formatting.format('foo', verbose=True), '\'foo\'')
    self.assertEqual(
        formatting.format('foo\'s\na', verbose=True), '"foo\'s\\na"')

    # Root indent has no impact on simple types.
    self.assertEqual(formatting.format(True, root_indent=4), 'True')
    self.assertEqual(formatting.format(1, root_indent=4), '1')
    self.assertEqual(formatting.format(1.0, root_indent=4), '1.0')
    self.assertEqual(formatting.format('foo', root_indent=4), '\'foo\'')
    self.assertEqual(
        formatting.format('foo\'s\na', root_indent=4), '"foo\'s\\na"')

  def test_complex_types(self):

    class CustomFormattable(formatting.Formattable):
      """Custom formattable."""

      def format(self, custom_param=None, **kwargs):
        return f'CustomFormattable({custom_param})'

    class A:
      pass

    self.assertEqual(
        formatting.format(
            {
                'a': CustomFormattable(),
                'b': {
                    'c': [1, 2, 3],
                    'd': ['foo', 'bar\na', 3, 4, 5]
                }
            },
            compact=True,
            custom_param='foo'),
        "{'a': CustomFormattable(foo), 'b': {'c': [1, 2, 3], "
        "'d': ['foo', 'bar\\na', 3, 4, 5]}}")

    self.assertEqual(
        formatting.format(
            {
                'a': A(),
                'b': {
                    'c': [1, 2, 3],
                    'd': ['foo', 'bar\na', 3, 4, 5]
                }
            },
            compact=False,
            list_wrap_threshold=15,
            strip_object_id=True),
        inspect.cleandoc("""{
          'a': A(...),
          'b': {
            'c': [1, 2, 3],
            'd': [
              'foo',
              'bar\\na',
              3,
              4,
              5
            ]
          }
        }"""))

  def test_include_exclude_keys(self):
    """Test format with excluded keys."""

    class A:
      pass

    class B(formatting.Formattable):
      """Custom formattable."""

      def format(
          self, custom_param=None,
          include_keys=None, exclude_keys=None, **kwargs):
        exclude_keys = exclude_keys or set()
        kv = dict(a=1, b=2, c=3)
        def _should_include(k):
          if include_keys:
            return k in include_keys
          return k not in exclude_keys
        kv_pairs = [(k, v, None) for k, v in kv.items() if _should_include(k)]
        return f'B({formatting.kvlist_str(kv_pairs, compact=True)})'

    self.assertEqual(
        formatting.format(B(), compact=False, include_keys=set(['a', 'c'])),
        'B(a=1, c=3)')
    self.assertEqual(
        formatting.format(B(), compact=False, exclude_keys=set(['a', 'c'])),
        'B(b=2)')
    self.assertEqual(
        formatting.format(
            {
                'a': A(),
                'b': B(),
                'c': {
                    'd': [1, 2, 3],
                }
            },
            compact=False,
            list_wrap_threshold=15,
            strip_object_id=True,
            # 'a' should be removed, but 'b.a', 'c.d' should be kept as they are
            # not at the top level.
            exclude_keys=set(['a', 'd'])),
        inspect.cleandoc("""{
          'b': B(a=1, b=2, c=3),
          'c': {
            'd': [1, 2, 3]
          }
        }"""))

  def test_custom_format(self):

    class A:

      def _repr_xml_(self):
        return '<a/>'

      def __repr__(self):
        return 'A()'

      def __str__(self):
        return 'AA()'

    self.assertEqual(formatting.format(A), str(A))
    self.assertEqual(formatting.format(A()), 'AA()')
    def _custom_format(v, root_indent):
      del root_indent
      f = getattr(v, '_repr_xml_', None)
      return f() if f else None

    self.assertEqual(
        formatting.format(A(), custom_format=_custom_format), '<a/>'
    )
    self.assertEqual(
        formatting.format(A(), compact=True),
        'A()'
    )
    self.assertEqual(
        formatting.format(A(), compact=True, custom_format=_custom_format),
        '<a/>'
    )
    self.assertEqual(
        formatting.format([A()], compact=True, custom_format=_custom_format),
        '[<a/>]'
    )
    self.assertEqual(
        formatting.format((A(), 1), compact=True, custom_format=_custom_format),
        '(<a/>, 1)'
    )
    self.assertEqual(
        formatting.format(
            dict(x=A()), compact=True, custom_format=_custom_format
        ),
        '{\'x\': <a/>}'
    )

  def test_markdown(self):

    class A(formatting.Formattable):
      def __init__(self, x):
        self.x = x

      def format(self, *args, **kwargs):
        del args, kwargs
        return 'A(' + formatting.format(self.x) + ')'

    with formatting.str_format(markdown=True):
      self.assertEqual(str(A(1)), '`A(1)`')
      self.assertEqual(str(A([A(1)])), '`A([A(1)])`')
      self.assertEqual(
          formatting.format([1], compact=True, markdown=True), '`[1]`'
      )

    self.assertEqual(
        formatting.format(
            [1, 2, 3], list_wrap_threshold=5, compact=False, markdown=True
        ),
        inspect.cleandoc("""
            ```
            [
              1,
              2,
              3
            ]
            ```
            """),
    )

  def test_max_len(self):
    self.assertEqual(
        formatting.format('foo', max_str_len=2), '\'fo...\''
    )
    self.assertEqual(
        formatting.format(b'bar', max_bytes_len=2), 'b\'ba...\''
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test for evolution recombinators."""

import unittest
import pyglove.core as pg
from pyglove.ext.evolution import recombinators


class RecombinationTest(unittest.TestCase):
  """Base class for recombination test."""

  def assert_eventual(
      self, recombinator, inputs, expected_outputs, max_attempts=200):
    """Asserts that the output should eventually cover all expected outputs."""
    expected_outputs = {tuple(sorted(o)): 0 for o in expected_outputs}
    matched = 0
    for _ in range(max_attempts):
      outputs = tuple(sorted(recombinator(inputs)))
      assert outputs in expected_outputs, outputs
      prev = expected_outputs[outputs]
      if prev == 0:
        matched += 1
      expected_outputs[outputs] = prev + 1
      if matched == len(expected_outputs):
        break
    remaining = [k for k, v in expected_outputs.items() if v == 0]
    assert not remaining, remaining


class UniformTest(RecombinationTest):
  """Tests for uniform recombinator."""

  def test_flat_search_space(self):
    dna_spec = pg.dna_spec(pg.List(
        [pg.oneof(range(10))]
        + [pg.floatv(0.1, 1.)]
        + [pg.manyof(2, range(10), False, False)]))
    r = recombinators.Uniform(seed=10)
    x = pg.DNA([1, 0.2, [1, 2]], spec=dna_spec)
    y = pg.DNA([3, 0.5, [4, 5]], spec=dna_spec)
    z = pg.DNA([2, 0.6, [3, 3]], spec=dna_spec)
    def _make_output(a, b, c, d):
      return (pg.DNA([a, b, [c, d]]),)
    self.assert_eventual(
        r, [x, y, z],
        [_make_output(*v) for v in pg.iter(pg.List([
            pg.oneof([1, 3, 2]), pg.oneof([0.2, 0.5, 0.6]),
            pg.oneof([1, 4, 3]), pg.oneof([2, 5, 3])]))],
        max_attempts=500)

  def test_hierarchical_search_space(self):
    dna_spec = pg.dna_spec(pg.manyof(3, [
        pg.oneof(['foo', pg.oneof([1, 2])]),
        pg.oneof(range(10)),
        'bar'
    ], False, False))
    r = recombinators.Uniform(seed=1)
    x = pg.DNA([(0, 0), 2, (1, 4)], spec=dna_spec)
    y = pg.DNA([2, (0, 1, 1), (1, 5)], spec=dna_spec)
    z = pg.DNA([(1, 3), 2, (0, 1, 0)], spec=dna_spec)
    def _make_output(a, b, c):
      return (pg.DNA([a, b, c]),)
    self.assert_eventual(
        r, [x, y, z],
        [_make_output(*v) for v in pg.iter(pg.List([
            pg.oneof([(0, 0), 2, (1, 3)]), pg.oneof([2, (0, 1, 1)]),
            pg.oneof([(1, 4), (1, 5), (0, 1, 0)])]))],
        max_attempts=500)

  def test_distinct_multi_choices(self):
    dna_spec = pg.dna_spec(pg.manyof(3, range(10), distinct=True))
    r = recombinators.Uniform(seed=1)
    x = pg.DNA([2, 1, 5], spec=dna_spec)
    y = pg.DNA([1, 0, 7], spec=dna_spec)
    z = pg.DNA([0, 1, 2], spec=dna_spec)
    self.assert_eventual(
        r, [x, y, z],
        [
            (pg.DNA([0, 1, 2]),),
            (pg.DNA([0, 1, 5]),),
            (pg.DNA([0, 1, 7]),),
            (pg.DNA([1, 0, 2]),),
            (pg.DNA([1, 0, 5]),),
            (pg.DNA([1, 0, 7]),),
            (pg.DNA([2, 0, 5]),),
            (pg.DNA([2, 0, 7]),),
            (pg.DNA([2, 1, 5]),),
            (pg.DNA([2, 1, 7]),),
        ])

  def test_sorted_multi_choices(self):
    dna_spec = pg.dna_spec(pg.manyof(3, range(10), distinct=False, sorted=True))
    r = recombinators.Uniform(seed=1)
    x = pg.DNA([1, 1, 5], spec=dna_spec)
    y = pg.DNA([0, 1, 2], spec=dna_spec)
    z = pg.DNA([2, 2, 2], spec=dna_spec)
    self.assert_eventual(
        r, [x, y, z],
        [
            (pg.DNA([0, 1, 2]),),
            (pg.DNA([0, 1, 5]),),
            (pg.DNA([0, 2, 2]),),
            (pg.DNA([0, 2, 5]),),
            (pg.DNA([1, 1, 2]),),
            (pg.DNA([1, 1, 5]),),
            (pg.DNA([1, 2, 2]),),
            (pg.DNA([1, 2, 5]),),
            (pg.DNA([2, 2, 2]),),
            (pg.DNA([2, 2, 5]),),
        ])

  def test_distinct_sorted_multi_choices(self):
    dna_spec = pg.dna_spec(pg.manyof(3, range(10), distinct=True, sorted=True))
    r = recombinators.Uniform(seed=1)
    x = pg.DNA([0, 1, 2], spec=dna_spec)
    y = pg.DNA([0, 2, 5], spec=dna_spec)
    z = pg.DNA([1, 5, 7], spec=dna_spec)
    self.assert_eventual(
        r, [x, y, z],
        [
            (pg.DNA([0, 1, 2]),),
            (pg.DNA([0, 1, 5]),),
            (pg.DNA([0, 1, 7]),),
            (pg.DNA([0, 2, 5]),),
            (pg.DNA([0, 2, 7]),),
            (pg.DNA([0, 5, 7]),),
            (pg.DNA([1, 2, 5]),),
            (pg.DNA([1, 2, 7]),),
            (pg.DNA([1, 5, 7]),),
        ])

  def test_distinct_multi_choices_without_rearrangement(self):
    dna_spec = pg.dna_spec(pg.permutate(range(4)))
    r = recombinators.Uniform(seed=1)
    x = pg.DNA([3, 2, 0, 1], spec=dna_spec)
    y = pg.DNA([0, 3, 1, 2], spec=dna_spec)
    self.assert_eventual(
        r, [x, y],
        [
            (pg.DNA([3, 2, 0, 1]),),
            (pg.DNA([0, 3, 1, 2]),),
        ])

  def test_carry_over_unrecombined_parts(self):
    dna_spec = pg.dna_spec(
        pg.Dict(x=pg.floatv(0.1, 1.0), y=pg.oneof(range(10))))
    r = recombinators.Uniform(
        where=lambda xs: [x for x in xs if isinstance(x, pg.geno.Float)],
        seed=1)
    x = pg.DNA([0.2, 1], spec=dna_spec)
    y = pg.DNA([0.5, 2], spec=dna_spec)
    z = pg.DNA([0.3, 7], spec=dna_spec)
    self.assert_eventual(
        r, [x, y, z],
        [
            (pg.DNA([0.2, 1]), pg.DNA([0.2, 2]), pg.DNA([0.2, 7])),
            (pg.DNA([0.5, 1]), pg.DNA([0.5, 2]), pg.DNA([0.5, 7])),
            (pg.DNA([0.3, 1]), pg.DNA([0.3, 2]), pg.DNA([0.3, 7])),
        ])

  def test_complex_case(self):
    dna_spec = pg.dna_spec(pg.Dict(
        x=pg.oneof(range(10), name='excluded'),
        y=pg.oneof([
            pg.manyof(2, [
                pg.oneof(range(10)),
                'foo',
                'bar'
            ]),
            1,
            2
        ]),
        z='constant'))

    r = recombinators.Uniform(
        where=lambda xs: [x for x in xs if x.name != 'excluded'], seed=1)
    x = pg.DNA([7, 1], spec=dna_spec)
    y = pg.DNA([5, (0, [(0, 4), 2])], spec=dna_spec)
    z = pg.DNA([3, (0, [1, (0, 7)])], spec=dna_spec)
    self.assert_eventual(
        r, [x, y, z],
        [
            (pg.DNA([7, 1]), pg.DNA([5, 1]), pg.DNA([3, 1])),
            (pg.DNA([7, (0, [(0, 4), 2])]),
             pg.DNA([5, (0, [(0, 4), 2])]),
             pg.DNA([3, (0, [(0, 4), 2])])),
            (pg.DNA([7, (0, [1, (0, 7)])]),
             pg.DNA([5, (0, [1, (0, 7)])]),
             pg.DNA([3, (0, [1, (0, 7)])])),
            (pg.DNA([7, (0, [1, 2])]),
             pg.DNA([5, (0, [1, 2])]),
             pg.DNA([3, (0, [1, 2])])),
        ], max_attempts=500)


class SampleTest(RecombinationTest):
  """Test sample recombinator."""

  def test_simple_case(self):
    dna_spec = pg.dna_spec(pg.Dict(
        x=pg.oneof(range(10)),
        y=pg.oneof(range(10))))

    r = recombinators.Sample(lambda parents: [0.0, 0.5, 0.5], seed=1)
    a = pg.DNA([1, 9], spec=dna_spec)
    b = pg.DNA([2, 3], spec=dna_spec)
    c = pg.DNA([5, 4], spec=dna_spec)
    self.assert_eventual(
        r, [a, b, c],
        [
            (pg.DNA([2, 3]),),
            (pg.DNA([2, 4]),),
            (pg.DNA([5, 3]),),
            (pg.DNA([5, 4]),),
        ])

  def test_complex_case(self):
    dna_spec = pg.dna_spec(pg.Dict(
        x=pg.oneof(range(10), name='excluded'),
        y=pg.oneof([
            pg.manyof(2, [
                pg.oneof(range(10)),
                'foo',
                'bar'
            ]),
            1,
            2
        ]),
        z='constant'))

    r = recombinators.Sample(
        weights=lambda parents: [0.0, 0.5, 0.5],
        where=lambda xs: [x for x in xs if x.name != 'excluded'], seed=1)
    a = pg.DNA([7, 1], spec=dna_spec)
    b = pg.DNA([5, (0, [(0, 4), 2])], spec=dna_spec)
    c = pg.DNA([3, (0, [1, (0, 7)])], spec=dna_spec)
    self.assert_eventual(
        r, [a, b, c],
        [
            # NOTE(daiyip): dict item y=2 shall never be sampled since
            # a's weight is 0.0.
            (pg.DNA([7, (0, [(0, 4), 2])]),
             pg.DNA([5, (0, [(0, 4), 2])]),
             pg.DNA([3, (0, [(0, 4), 2])])),
            (pg.DNA([7, (0, [1, (0, 7)])]),
             pg.DNA([5, (0, [1, (0, 7)])]),
             pg.DNA([3, (0, [1, (0, 7)])])),
            (pg.DNA([7, (0, [1, 2])]),
             pg.DNA([5, (0, [1, 2])]),
             pg.DNA([3, (0, [1, 2])])),
        ], max_attempts=500)


class AverageTest(RecombinationTest):
  """Test average recombinator."""

  def test_floats_only(self):
    dna_spec = pg.dna_spec(
        pg.Dict(x=pg.floatv(0.1, 0.5), y=pg.floatv(0.0, 1.0)))

    r = recombinators.Average()
    x = pg.DNA([0.1, 0.5], spec=dna_spec)
    y = pg.DNA([0.5, 0.2], spec=dna_spec)
    z = pg.DNA([0.3, 0.8], spec=dna_spec)
    self.assert_eventual(
        r, [x, y, z],
        [
            (pg.DNA([0.3, 0.5]),)
        ],)

  def test_carry_over_unrecombined_parts(self):
    dna_spec = pg.dna_spec(
        pg.Dict(x=pg.floatv(0.1, 0.5, name='excluded'), y=pg.floatv(0.0, 1.0)))

    r = recombinators.Average(
        where=lambda xs: [x for x in xs if x.name != 'excluded'])
    x = pg.DNA([0.1, 0.5], spec=dna_spec)
    y = pg.DNA([0.5, 0.2], spec=dna_spec)
    z = pg.DNA([0.3, 0.8], spec=dna_spec)
    self.assert_eventual(
        r, [x, y, z],
        [(pg.DNA([0.1, 0.5]), pg.DNA([0.5, 0.5]), pg.DNA([0.3, 0.5]))])


class WeightedAverageTest(RecombinationTest):
  """Test weighted average recombinator."""

  def test_floats_only(self):
    dna_spec = pg.dna_spec(
        pg.Dict(x=pg.floatv(0.0, 1.0), y=pg.floatv(0.0, 1.0)))

    r = recombinators.WeightedAverage(lambda x: [0.2, 0.2, 0.6])
    x = pg.DNA([0.1, 0.5], spec=dna_spec)
    y = pg.DNA([0.5, 0.2], spec=dna_spec)
    z = pg.DNA([0.3, 0.8], spec=dna_spec)
    self.assert_eventual(
        r, [x, y, z],
        [
            (pg.DNA([0.3, 0.62]),)
        ],)

  def test_carry_over_unrecombined_parts(self):
    dna_spec = pg.dna_spec(
        pg.Dict(x=pg.floatv(0.1, 0.5, name='excluded'), y=pg.floatv(0.0, 1.0)))

    r = recombinators.WeightedAverage(
        lambda x: [0.2, 0.2, 0.6],
        where=lambda xs: [x for x in xs if x.name != 'excluded'])
    x = pg.DNA([0.1, 0.5], spec=dna_spec)
    y = pg.DNA([0.5, 0.2], spec=dna_spec)
    z = pg.DNA([0.3, 0.8], spec=dna_spec)
    self.assert_eventual(
        r, [x, y, z],
        [(pg.DNA([0.1, 0.62]), pg.DNA([0.5, 0.62]), pg.DNA([0.3, 0.62]))])


class KPointTest(RecombinationTest):
  """Test KPoint recombinator."""

  def test_single_point(self):
    dna_spec = pg.dna_spec(
        pg.List([pg.oneof(range(100))] * 4
                + [pg.floatv(0., 1.)]))

    r = recombinators.KPoint(1, seed=2)
    x = pg.DNA([0, 1, 3, 1, 0.5], spec=dna_spec)
    y = pg.DNA([1, 2, 5, 4, 0.3], spec=dna_spec)
    self.assert_eventual(
        r, [x, y],
        [
            (pg.DNA([0, 2, 5, 4, 0.3]), pg.DNA([1, 1, 3, 1, 0.5])),
            (pg.DNA([0, 1, 5, 4, 0.3]), pg.DNA([1, 2, 3, 1, 0.5])),
            (pg.DNA([0, 1, 3, 4, 0.3]), pg.DNA([1, 2, 5, 1, 0.5])),
            (pg.DNA([0, 1, 3, 1, 0.3]), pg.DNA([1, 2, 5, 4, 0.5])),
        ])

  def test_two_point(self):
    dna_spec = pg.dna_spec(pg.List([pg.oneof(range(100))] * 4))

    r = recombinators.KPoint(2, seed=2)
    x = pg.DNA([0, 1, 3, 1], spec=dna_spec)
    y = pg.DNA([1, 2, 3, 4], spec=dna_spec)
    self.assert_eventual(
        r, [x, y],
        [
            (pg.DNA([0, 2, 3, 1]), pg.DNA([1, 1, 3, 4])),
            (pg.DNA([0, 1, 3, 1]), pg.DNA([1, 2, 3, 4])),
        ])

  def test_k_points(self):
    dna_spec = pg.dna_spec(pg.List([pg.oneof(range(100))] * 4))

    # Downgrade into alternating position.
    r = recombinators.KPoint(5, seed=2)
    x = pg.DNA([0, 1, 3, 1], spec=dna_spec)
    y = pg.DNA([1, 2, 5, 4], spec=dna_spec)
    self.assert_eventual(
        r, [x, y],
        [
            (pg.DNA([0, 2, 3, 4]), pg.DNA([1, 1, 5, 1])),
        ])

  def test_independent_multi_choice(self):
    dna_spec = pg.dna_spec(pg.Dict(
        x=[pg.oneof(range(10))] * 2,
        y=pg.manyof(2, range(10), distinct=False)))

    r = recombinators.KPoint(2, seed=2)
    x = pg.DNA([0, 1, [3, 1]], spec=dna_spec)
    y = pg.DNA([1, 2, [5, 4]], spec=dna_spec)
    self.assert_eventual(
        r, [x, y],
        [
            (pg.DNA([0, 2, [3, 1]]), pg.DNA([1, 1, [5, 4]])),
            (pg.DNA([0, 2, [5, 1]]), pg.DNA([1, 1, [3, 4]])),
            (pg.DNA([0, 1, [5, 1]]), pg.DNA([1, 2, [3, 4]])),
        ])

  def test_distinct_multi_choice(self):
    dna_spec = pg.dna_spec(pg.Dict(
        x=[pg.oneof(range(10))] * 2,
        y=pg.manyof(3, range(10), distinct=True)))

    r = recombinators.KPoint(2, seed=2)
    x = pg.DNA([0, 1, [3, 1, 2]], spec=dna_spec)
    y = pg.DNA([1, 2, [1, 3, 4]], spec=dna_spec)
    self.assert_eventual(
        r, [x, y],
        [
            (pg.DNA([0, 2, [3, 1, 2]]), pg.DNA([1, 1, [1, 3, 4]])),
        ])

  def test_sorted_multi_choice(self):
    dna_spec = pg.dna_spec(pg.Dict(
        x=[pg.oneof(range(10))] * 2,
        y=pg.manyof(3, range(10), sorted=True)))

    r = recombinators.KPoint(2, seed=2)
    x = pg.DNA([0, 1, [1, 2, 3]], spec=dna_spec)
    y = pg.DNA([1, 2, [1, 3, 4]], spec=dna_spec)
    self.assert_eventual(
        r, [x, y],
        [
            (pg.DNA([0, 2, [1, 2, 3]]), pg.DNA([1, 1, [1, 3, 4]])),
        ])


class SegmentedTest(RecombinationTest):
  """Test for alternating position recombinator."""

  def test_basics(self):
    dna_spec = pg.dna_spec(pg.List([pg.oneof(range(100))] * 4))
    r = recombinators.Segmented(lambda xs: [1, 3])
    x = pg.DNA([0, 1, 3, 1], spec=dna_spec)
    y = pg.DNA([1, 2, 5, 4], spec=dna_spec)
    self.assert_eventual(
        r, [x, y],
        [
            (pg.DNA([0, 2, 5, 1]), pg.DNA([1, 1, 3, 4]))
        ])


class PartiallyMappedTest(RecombinationTest):
  """Test for partially mapped crossover (PMX)."""

  def test_simple_case(self):
    dna_spec = pg.dna_spec(pg.permutate(range(5)))
    r = recombinators.PartiallyMapped(seed=3)
    x = pg.DNA([0, 1, 2, 4, 3], spec=dna_spec)
    y = pg.DNA([1, 2, 3, 4, 0], spec=dna_spec)
    self.assert_eventual(
        r, [x, y],
        [
            (pg.DNA([0, 2, 3, 4, 1]), pg.DNA([3, 1, 2, 4, 0]))
        ])

  def test_non_matching_multi_choices(self):
    dna_spec = pg.dna_spec(
        pg.oneof([
            pg.permutate(range(5)),
            pg.permutate(range(5))
        ]))

    r = recombinators.PartiallyMapped(seed=3)
    x = pg.DNA((0, [0, 1, 2, 4, 3]), spec=dna_spec)
    y = pg.DNA((1, [1, 2, 3, 4, 0]), spec=dna_spec)

    # Output the parents since two multi-choices are active at different branch.
    self.assert_eventual(
        r, [x, y],
        [
            (pg.DNA((0, [0, 1, 2, 4, 3])), pg.DNA((1, [1, 2, 3, 4, 0]))),
        ])

  def test_multi_choices_with_different_decisions(self):
    dna_spec = pg.dna_spec(pg.Dict(
        x=pg.oneof([1, 2, 3]),
        y=pg.permutate(range(3)),
        z=pg.floatv(0., 1.)))

    r = recombinators.PartiallyMapped(seed=3)
    x = pg.DNA([0, [0, 1, 2], 0.2], spec=dna_spec)
    y = pg.DNA([1, [1, 2, 0], 0.5], spec=dna_spec)

    # The output is a product of non-permutated parts of the parents and the
    # permutated parts.
    self.assert_eventual(
        r, [x, y],
        [
            (pg.DNA([0, [0, 1, 2], 0.2]),
             pg.DNA([1, [0, 1, 2], 0.5]),
             pg.DNA([1, [1, 2, 0], 0.5]),
             pg.DNA([0, [1, 2, 0], 0.2]))
        ])

  def test_nested_matching_multi_choices(self):
    dna_spec = pg.dna_spec(pg.oneof([
        1,
        pg.permutate(range(5))
    ]))

    r = recombinators.PartiallyMapped(seed=3)
    x = pg.DNA((1, [0, 1, 2, 4, 3]), spec=dna_spec)
    y = pg.DNA((1, [1, 2, 3, 4, 0]), spec=dna_spec)
    self.assert_eventual(
        r, [x, y],
        [
            (pg.DNA((1, [1, 2, 3, 4, 0])), pg.DNA((1, [0, 1, 2, 4, 3])))
        ])

  def test_where_clause(self):
    dna_spec = pg.dna_spec(pg.List([
        pg.permutate(range(5), name='a'),
        pg.permutate(range(5), name='b'),
        pg.permutate(range(5), name='c')
    ]))

    r = recombinators.PartiallyMapped(
        where=lambda xs: [x for x in xs if x.name != 'b'], seed=5)
    x = pg.DNA([[0, 1, 2, 3, 4],
                [1, 2, 3, 4, 0],
                [2, 3, 4, 0, 1]], spec=dna_spec)
    y = pg.DNA([[3, 4, 0, 1, 2],
                [4, 0, 1, 2, 3],
                [0, 1, 2, 3, 4]], spec=dna_spec)
    self.assert_eventual(
        r, [x, y],
        [
            # Crossover happens on the 2rd column. Without the `where`
            # statement it will happen on the 1st and 3rd column.
            (pg.DNA([[3, 4, 0, 1, 2], [4, 0, 1, 2, 3], [2, 3, 0, 1, 4]]),
             pg.DNA([[2, 3, 0, 1, 4], [1, 2, 3, 4, 0], [2, 3, 4, 0, 1]]),
             pg.DNA([[0, 1, 2, 3, 4], [1, 2, 3, 4, 0], [2, 3, 0, 1, 4]]),
             pg.DNA([[3, 4, 0, 1, 2], [4, 0, 1, 2, 3], [0, 1, 4, 2, 3]]),
             pg.DNA([[0, 1, 2, 3, 4], [1, 2, 3, 4, 0], [0, 1, 4, 2, 3]]),
             pg.DNA([[1, 4, 2, 3, 0], [4, 0, 1, 2, 3], [0, 1, 2, 3, 4]]),
             pg.DNA([[2, 3, 0, 1, 4], [4, 0, 1, 2, 3], [0, 1, 2, 3, 4]]),
             pg.DNA([[1, 4, 2, 3, 0], [1, 2, 3, 4, 0], [2, 3, 4, 0, 1]]))
        ])


class OrderCrossoverTest(RecombinationTest):
  """Test for order crossover (OX)."""

  def test_correctness(self):
    """Refer to https://www.hindawi.com/journals/cin/2017/7430125."""
    r = recombinators.Order()
    self.assertEqual(
        r.order_crossover([[3, 4, 8, 2, 7, 1, 6, 5],
                           [4, 2, 5, 1, 6, 8, 3, 7]], 3, 6),
        [[4, 2, 7, 1, 6, 8, 5, 3], [5, 6, 8, 2, 7, 1, 3, 4]])

  def test_simple_case(self):
    dna_spec = pg.dna_spec(pg.permutate(range(5)))
    r = recombinators.Order(seed=4)
    x = pg.DNA([0, 1, 2, 4, 3], spec=dna_spec)
    y = pg.DNA([1, 2, 3, 4, 0], spec=dna_spec)
    self.assert_eventual(
        r, [x, y],
        [
            (pg.DNA([1, 2, 4, 3, 0]), pg.DNA([2, 1, 3, 4, 0]))
        ])


class CycleCrossoverTest(RecombinationTest):
  """Test for cycle crossover (CX)."""

  def test_correctness(self):
    """Refer to https://www.hindawi.com/journals/cin/2017/7430125."""
    r = recombinators.Cycle(seed=4)
    self.assertEqual(
        r.permutate(None, [[1, 2, 3, 4, 5, 6, 7, 8],
                           [8, 5, 2, 1, 3, 6, 4, 7]]),
        [[1, 5, 2, 4, 3, 6, 7, 8], [8, 2, 3, 1, 5, 6, 4, 7]])

  def test_simple_case(self):
    dna_spec = pg.dna_spec(pg.permutate(range(5)))
    r = recombinators.Cycle(seed=4)
    x = pg.DNA([0, 1, 2, 3, 4], spec=dna_spec)
    y = pg.DNA([1, 0, 3, 4, 2], spec=dna_spec)
    self.assert_eventual(
        r, [x, y],
        [
            (pg.DNA([0, 1, 3, 4, 2]), pg.DNA([1, 0, 2, 3, 4]))
        ])


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Boilerplate class from symbolic object."""

import copy
import inspect

from typing import Any, List, Optional, Type

from pyglove.core import object_utils
from pyglove.core import typing as pg_typing
from pyglove.core.symbolic import flags
from pyglove.core.symbolic import object as pg_object


def boilerplate_class(
    cls_name: str,
    value: pg_object.Object,
    init_arg_list: Optional[List[str]] = None,
    **kwargs) -> Type[pg_object.Object]:
  """Create a boilerplate class using a symbolic object.

  As the name indicates, a boilerplate class is a class that can be used
  as a boilerplate to create object.

  Implementation-wise it's a class that extends the type of input value, while
  setting the default values of its (inherited) schema using the value from
  input.

  An analogy to boilerplate class is prebound function.

    For example::

      # A regular function: correspond to a pg.Object subclass.
      def f(a, b, c)
        return a + b + c

      # A partially bound function: correspond to a boilerplate class created
      # from a partially bound object.
      def g(c):
        return f(1, 2, c)

      # A fully bound function: correspond to a boilerplate class created from
      # a fully bound object.
      def h():
        return f(1, 2, 3)

  Boilerplate class can be created with a value that is fully bound
  (like function `h` above), or partially bound (like function `g` above).
  Since boilerplate class extends the type of the input, we can rebind members
  of its instances as we modify the input.

  Here are a few examples::

    @pg.members([
      ('a', pg.typing.Str(), 'Field A.'),
      ('b', pg.typing.Int(), 'Field B.'),
    ])
    class A(pg.Object):
      pass

    A1 = pg.boilerplate_class('A1', A.partial(a='foo'))
    assert A1(b=1) == A(a='foo', b=1)

    A2 = pg.boilerplate_class('A2', A(a='bar', b=2))
    assert A2() == A(a='bar', b=2)

  Args:
    cls_name: Name of the boilerplate class.
    value: Value that is used as the default value of the boilerplate class.
    init_arg_list: An optional list of strings as __init__ positional arguments
      names.
    **kwargs: Keyword arguments for infrequently used options. Acceptable
      keywords are:  * `serialization_key`: An optional string to be used as the
      serialization key for the class during `sym_jsonify`. If None,
      `cls.__type_name__` will be used. This is introduced for scenarios when we
      want to relocate a class, before the downstream can recognize the new
      location, we need the class to serialize it using previous key. *
      `additional_keys`: An optional list of strings as additional keys to
      deserialize an object of the registered class. This can be useful when we
      need to relocate or rename the registered class while being able to load
      existing serialized JSON values.

  Returns:
    A class which extends the input value's type, with its schema's default
      values set from the input value.

  Raises:
    TypeError: Keyword argumment provided is not supported.
  """
  if not isinstance(value, pg_object.Object):
    raise ValueError('Argument \'value\' must be an instance of '
                     'symbolic.Object subclass.')

  serialization_key = kwargs.pop('serialization_key', None)
  additional_keys = kwargs.pop('additional_keys', None)
  if kwargs:
    raise TypeError(
        f'Unsupported keyword arguments: {list(kwargs.keys())!r}.')

  base_cls = value.__class__

  class _BoilerplateClass(base_cls):
    """Boilerplate class."""
    auto_register = False

    is_boilerplate = True

  caller_module = inspect.getmodule(inspect.stack()[1][0])
  cls_module = caller_module.__name__ if caller_module else '__main__'
  cls = _BoilerplateClass
  cls.__name__ = cls_name
  cls.__qualname__ = cls.__qualname__.replace(
      'boilerplate_class.<locals>._BoilerplateClass', cls_name)
  cls.__module__ = cls_module

  # Enable automatic registration for subclass.
  cls.auto_register = True

  allow_partial = value.allow_partial
  def _freeze_field(path: object_utils.KeyPath,
                    field: pg_typing.Field,
                    value: Any) -> Any:
    # We do not do validation since Object is already in valid form.
    del path
    if not isinstance(field.key, pg_typing.ListKey):
      # Recursively freeze dict field.
      if isinstance(field.value, pg_typing.Dict) and field.value.schema:
        field.value.schema.apply(
            value, allow_partial=allow_partial, child_transform=_freeze_field)
        field.value.set_default(value)
        if all(f.frozen for f in field.value.schema.values()):
          field.value.freeze()
      else:
        if value != pg_typing.MISSING_VALUE:
          field.value.freeze(copy.deepcopy(value), apply_before_use=False)
        else:
          field.value.set_default(
              pg_typing.MISSING_VALUE, use_default_apply=False)
    return value

  # NOTE(daiyip): we call `cls.__schema__.apply` to freeze fields that have
  # default values.
  with flags.allow_writable_accessors():
    cls.__schema__.apply(
        value._sym_attributes,  # pylint: disable=protected-access
        allow_partial=allow_partial,
        child_transform=_freeze_field,
    )
  cls.__schema__.metadata['init_arg_list'] = init_arg_list
  cls.apply_schema(cls.__schema__)
  cls.register_for_deserialization(serialization_key, additional_keys)
  return cls



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for formatting objects."""

import abc
import enum
import io
import sys
from typing import Any, Callable, ContextManager, Dict, List, Optional, Sequence, Set, Tuple
from pyglove.core.object_utils import thread_local


_TLS_STR_FORMAT_KWARGS = '_str_format_kwargs'
_TLS_REPR_FORMAT_KWARGS = '_repr_format_kwargs'


CustomFormatFn = Callable[
    [
        Any,    # Value to format.
        int,    # Root indent.
    ],
    # Returns a string or None. If None, the default `pg.format` will be used.
    Optional[str]
]


def str_format(**kwargs) -> ContextManager[Dict[str, Any]]:
  """Context manager for setting the default format kwargs for __str__."""
  return thread_local.thread_local_arg_scope(_TLS_STR_FORMAT_KWARGS, **kwargs)


def repr_format(**kwargs) -> ContextManager[Dict[str, Any]]:
  """Context manager for setting the default format kwargs for __repr__."""
  return thread_local.thread_local_arg_scope(_TLS_REPR_FORMAT_KWARGS, **kwargs)


class Formattable(metaclass=abc.ABCMeta):
  """Interface for classes whose instances can be pretty-formatted.

  This interface overrides the default ``__repr__`` and ``__str__`` method, thus
  all ``Formattable`` objects can be printed nicely.

  All symbolic types implement this interface.
  """

  # Additional format keyword arguments for `__str__`.
  __str_format_kwargs__ = dict(compact=False, verbose=True)

  # Additional format keyword arguments for `__repr__`.
  __repr_format_kwargs__ = dict(compact=True)

  @abc.abstractmethod
  def format(self,
             compact: bool = False,
             verbose: bool = True,
             root_indent: int = 0,
             **kwargs) -> str:
    """Formats this object into a string representation.

    Args:
      compact: If True, this object will be formatted into a single line.
      verbose: If True, this object will be formatted with verbosity.
        Subclasses should define `verbosity` on their own.
      root_indent: The start indent level for this object if the output is a
        multi-line string.
      **kwargs: Subclass specific keyword arguments.

    Returns:
      A string of formatted object.
    """

  def __str__(self) -> str:
    """Returns the full (maybe multi-line) representation of this object."""
    # NOTE(daiyip): we delegate the formatting logic to `pg.format` instead of
    # `Formattable.format` as `pg.format` could add common functionalities such
    # as `markdown` support.
    return format(self, **self.__str_kwargs__())

  def __str_kwargs__(self) -> Dict[str, Any]:
    """Returns the default format kwargs for __str__."""
    kwargs = dict(self.__str_format_kwargs__)
    kwargs.update(thread_local.thread_local_kwargs(_TLS_STR_FORMAT_KWARGS))
    return kwargs

  def __repr__(self) -> str:
    """Returns a single-line representation of this object."""
    # NOTE(daiyip): we delegate the formatting logic to `pg.format` instead of
    # `Formattable.format` as `pg.format` could add common functionalities such
    # as `markdown` support.
    return format(self, **self.__repr_kwargs__())

  def __repr_kwargs__(self) -> Dict[str, Any]:
    """Returns the default format kwargs for __repr__."""
    kwargs = dict(self.__repr_format_kwargs__)
    kwargs.update(thread_local.thread_local_kwargs(_TLS_REPR_FORMAT_KWARGS))
    return kwargs


class RawText(Formattable):
  """Raw text."""

  def __init__(self, text: str):
    self.text = text

  def format(self, *args, **kwargs):
    del args, kwargs
    return self.text

  def __eq__(self, other: Any) -> bool:
    if isinstance(other, RawText):
      return self.text == other.text
    elif isinstance(other, str):
      return self.text == other
    return False

  def __ne__(self, other: Any) -> bool:
    return not self.__eq__(other)


class BracketType(enum.IntEnum):
  """Bracket types used for complex type formatting."""
  # Round bracket.
  ROUND = 0

  # Square bracket.
  SQUARE = 1

  # Curly bracket.
  CURLY = 2


_BRACKET_CHARS = [
    ('(', ')'),
    ('[', ']'),
    ('{', '}'),
]


def bracket_chars(bracket_type: BracketType) -> Tuple[str, str]:
  """Gets bracket character."""
  return _BRACKET_CHARS[int(bracket_type)]


def kvlist_str(
    kvlist: List[Tuple[str, Any, Any]],
    compact: bool = True,
    verbose: bool = False,
    root_indent: int = 0,
    *,
    label: Optional[str] = None,
    bracket_type: BracketType = BracketType.ROUND,
    custom_format: Optional[CustomFormatFn] = None,
    **kwargs,
) -> str:
  """Formats a list key/value pairs into a comma delimited string.

  Args:
    kvlist: List of tuples in format of
      (key, value, default_value or a tuple of default values)
    compact: If True, format value in kvlist in compact form.
    verbose: If True, format value in kvlist in verbose.
    root_indent: The indent should be applied for values in kvlist if they are
      multi-line.
    label: (Optional) If not None, add label to brace all kv pairs.
    bracket_type: Bracket type used for embracing the kv pairs. Applicable only
      when `name` is not None.
    custom_format: An optional custom format function, which will be applied to
      each value (and child values) in kvlist. If the function returns None, it
      will fall back to the default `pg.format`.
    **kwargs: Keyword arguments that will be passed through unto child
      ``Formattable`` objects.
  Returns:
    A formatted string from a list of key/value pairs delimited by comma.
  """
  s = io.StringIO()
  is_first = True
  bracket_start, bracket_end = bracket_chars(bracket_type)

  child_indent = (root_indent + 1) if label else root_indent
  body = io.StringIO()

  for k, v, d in kvlist:
    if isinstance(d, tuple):
      include_pair = True
      for sd in d:
        if sd == v:
          include_pair = False
          break
    else:
      include_pair = v != d
    if include_pair:
      if not is_first:
        body.write(',')
        body.write(' ' if compact else '\n')
      v = format(
          v, compact=compact,
          verbose=verbose,
          root_indent=child_indent,
          custom_format=custom_format,
          **kwargs
      )
      if not compact:
        body.write(_indent('', child_indent))
      if k:
        body.write(f'{k}={str_ext(v, custom_format, child_indent)}')
      else:
        body.write(str_ext(v, custom_format, child_indent))
      is_first = False

  if label and not is_first and not compact:
    body.write('\n')

  body = body.getvalue()
  if label is None:
    return body
  else:
    s.write(label)
    s.write(bracket_start)
    if body:
      if not compact:
        s.write('\n')
      s.write(body)
      if not compact:
        s.write(_indent('', root_indent))
    s.write(bracket_end)
    return s.getvalue()


def quote_if_str(value: Any) -> Any:
  """Quotes the value if it is a str."""
  if isinstance(value, str):
    return repr(value)
  return value


def comma_delimited_str(value_list: Sequence[Any]) -> str:
  """Gets comma delimited string."""
  return ', '.join(str(quote_if_str(v)) for v in value_list)


def auto_plural(
    number: int, singular: str, plural: Optional[str] = None) -> str:
  """Use singular form if number is 1, otherwise use plural form."""
  if plural is None:
    plural = singular + 's'
  return singular if number == 1 else plural


def format(   # pylint: disable=redefined-builtin
    value: Any,
    compact: bool = False,
    verbose: bool = True,
    root_indent: int = 0,
    list_wrap_threshold: int = 80,
    strip_object_id: bool = False,
    include_keys: Optional[Set[str]] = None,
    exclude_keys: Optional[Set[str]] = None,
    markdown: bool = False,
    max_str_len: Optional[int] = None,
    max_bytes_len: Optional[int] = None,
    *,
    custom_format: Optional[CustomFormatFn] = None,
    **kwargs,
) -> str:
  """Formats a (maybe) hierarchical value with flags.

  Args:
    value: The value to format.
    compact: If True, this object will be formatted into a single line.
    verbose: If True, this object will be formatted with verbosity. Subclasses
      should define `verbosity` on their own.
    root_indent: The start indent level for this object if the output is a
      multi-line string.
    list_wrap_threshold: A threshold in number of characters for wrapping a list
      value in a single line.
    strip_object_id: If True, format object as '<class-name>(...)' other than
      'object at <address>'.
    include_keys: A set of keys to include from the top-level dict or object.
    exclude_keys: A set of keys to exclude from the top-level dict or object.
      Applicable only when `include_keys` is set to None.
    markdown: If True, use markdown notion to quote the formatted object.
    max_str_len: The max length of the string to be formatted. If the string is
      longer than this length, it will be truncated.
    max_bytes_len: The max length of the bytes to be formatted. If the bytes is
      longer than this length, it will be truncated.
    custom_format: An optional custom format function, which will be applied to
      each value (and child values) in kvlist. If the function returns None, it
      will fall back to the default `pg.format`.
    **kwargs: Keyword arguments that will be passed through unto child
      ``Formattable`` objects.

  Returns:
    A string representation for `value`.
  """
  # We allow custom_format to intercept the entire value if it's present.
  if custom_format is not None:
    result = custom_format(value, root_indent)
    if result is not None:
      return maybe_markdown_quote(result, markdown)

  exclude_keys = exclude_keys or set()

  def _should_include_key(key: str) -> bool:
    if include_keys:
      return key in include_keys
    return key not in exclude_keys

  def _format_child(v):
    return format(
        v,
        compact=compact,
        verbose=verbose,
        root_indent=root_indent + 1,
        list_wrap_threshold=list_wrap_threshold,
        strip_object_id=strip_object_id,
        max_str_len=max_str_len,
        max_bytes_len=max_bytes_len,
        custom_format=custom_format,
        **kwargs
    )

  # `markdown` is only applied at the outter most level,
  # so we disable markdown format in the sub-tree for `str` and `repr`.
  with str_format(markdown=False), repr_format(markdown=False):
    if isinstance(value, Formattable):
      s = value.format(
          compact=compact,
          verbose=verbose,
          root_indent=root_indent,
          list_wrap_threshold=list_wrap_threshold,
          strip_object_id=strip_object_id,
          include_keys=include_keys,
          exclude_keys=exclude_keys,
          max_str_len=max_str_len,
          max_bytes_len=max_bytes_len,
          custom_format=custom_format,
          **kwargs
      )
    elif isinstance(value, (list, tuple)):
      # Always try compact representation if length is not too long.
      open_bracket, close_bracket = bracket_chars(
          BracketType.SQUARE if isinstance(value, list) else BracketType.ROUND)
      s = [open_bracket]
      s.append(', '.join([_format_child(elem) for elem in value]))
      s.append(close_bracket)
      s = [''.join(s)]
      if not compact and len(s[-1]) > list_wrap_threshold:
        s = [f'{open_bracket}\n']
        s.append(',\n'.join([
            _indent(_format_child(elem), root_indent + 1)
            for elem in value
        ]))
        s.append('\n')
        s.append(_indent(close_bracket, root_indent))
    elif isinstance(value, dict):
      if compact or not value:
        s = ['{']
        s.append(', '.join([
            f'{k!r}: {_format_child(v)}'
            for k, v in value.items() if _should_include_key(k)
        ]))
        s.append('}')
      else:
        s = ['{\n']
        s.append(',\n'.join([
            _indent(f'{k!r}: {_format_child(v)}', root_indent + 1)
            for k, v in value.items() if _should_include_key(k)
        ]))
        s.append('\n')
        s.append(_indent('}', root_indent))
    else:
      if isinstance(value, str):
        if max_str_len is not None and len(value) > max_str_len:
          value = value[:max_str_len] + '...'
        s = [repr_ext(value, custom_format, root_indent)]
      elif isinstance(value, bytes):
        if max_bytes_len is not None and len(value) > max_bytes_len:
          value = value[:max_bytes_len] + b'...'
        s = [repr_ext(value, custom_format, root_indent)]
      else:
        s = [repr_ext(value, custom_format, root_indent)
             if compact else str_ext(value, custom_format, root_indent)]
        if strip_object_id and 'object at 0x' in s[-1]:
          s = [f'{value.__class__.__name__}(...)']
  return maybe_markdown_quote(''.join(s), markdown)


def _maybe_custom_format(
    v: Any,
    default_fn: Callable[[Any], str],
    custom_format: Optional[CustomFormatFn],
    root_indent: int,
) -> str:
  if custom_format is None:
    return default_fn(v)
  x = custom_format(v, root_indent)
  if x is None:
    return default_fn(v)
  return x


def str_ext(
    v: Any,
    custom_format: Optional[CustomFormatFn] = None,
    root_indent: int = 0,
) -> str:
  """"str operator with special format support."""
  return _maybe_custom_format(v, str, custom_format, root_indent)


def repr_ext(
    v: Any,
    custom_format: Optional[CustomFormatFn] = None,
    root_indent: int = 0,
) -> str:
  """repr operator with special format support."""
  return _maybe_custom_format(v, repr, custom_format, root_indent)


def maybe_markdown_quote(s: str, markdown: bool = True) -> str:
  """Maybe quote the formatted string with markdown."""
  if not markdown:
    return s
  if '\n' not in s:
    return f'`{s}`'
  else:
    return f'```\n{s}\n```'


def printv(v: Any, **kwargs):
  """Prints formatted value."""
  fs = kwargs.pop('file', sys.stdout)
  print(format(v, **kwargs), file=fs)


def _indent(text: str, indent: int) -> str:
  return ' ' * 2 * indent + text



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Categorical hyper primitives."""

import numbers
import typing
from typing import Any, Callable, Iterable, List, Optional, Tuple, Union

from pyglove.core import geno
from pyglove.core import object_utils
from pyglove.core import symbolic
from pyglove.core import typing as pg_typing
from pyglove.core.hyper import base
from pyglove.core.hyper import object_template


@symbolic.members([
    ('num_choices', pg_typing.Int(min_value=0).noneable(),
     ('Number of choices to make. If set to None, any number of choices is '
      'acceptable.')),
    ('candidates', pg_typing.List(pg_typing.Any()),
     ('Candidate values, which may contain nested hyper values.'
      'Candidate can customize its display value (literal) by implementing the '
      '`pg.Formattable` interface.')),
    ('choices_distinct', pg_typing.Bool(True), 'Whether choices are distinct.'),
    ('choices_sorted', pg_typing.Bool(False), 'Whether choices are sorted.'),
    ('where', pg_typing.Callable([pg_typing.Object(base.HyperPrimitive)],
                                 returns=pg_typing.Bool()).noneable(),
     ('Callable object to filter nested hyper values. If None, all nested '
      'hyper value will be included in the encoding/decoding process. '
      'Otherwise only the hyper values on which `where` returns True will be '
      'included. `where` can be useful to partition a search space into '
      'separate optimization processes. '
      'Please see `ObjectTemplate` docstr for details.'))
])
class Choices(base.HyperPrimitive):
  """Categorical choices from a list of candidates.

  Example::

    # A single categorical choice:
    v = pg.oneof([1, 2, 3])

    # A multiple categorical choice as a list:
    vs = pg.manyof(2, [1, 2, 3])

    # A hierarchical categorical choice:
    v2 = pg.oneof([
        'foo',
        'bar',
        pg.manyof(2, [1, 2, 3])
    ])

  See also:

    * :class:`pyglove.hyper.OneOf`
    * :class:`pyglove.hyper.ManyOf`
    * :func:`pyglove.oneof`
    * :func:`pyglove.manyof`
    * :func:`pyglove.permutate`
  """

  def _on_bound(self):
    """On members are bound."""
    super()._on_bound()
    if self.num_choices > len(self.candidates) and self.choices_distinct:
      raise ValueError(
          f'{len(self.candidates)} candidates cannot produce '
          f'{self.num_choices} distinct choices.')
    self._candidate_templates = [
        object_template.ObjectTemplate(c, where=self.where)
        for c in self.candidates]
    # ValueSpec for candidate.
    self._value_spec = None

  def _update_children_paths(
      self, old_path: object_utils.KeyPath, new_path: object_utils.KeyPath):
    """Customized logic to update children paths."""
    super()._update_children_paths(old_path, new_path)
    for t in self._candidate_templates:
      t.root_path = self.sym_path

  @property
  def candidate_templates(self):
    """Returns candidate templates."""
    return self._candidate_templates

  @property
  def is_leaf(self) -> bool:
    """Returns whether this is a leaf node."""
    for t in self._candidate_templates:
      if not t.is_constant:
        return False
    return True

  def dna_spec(self,
               location: Optional[object_utils.KeyPath] = None) -> geno.Choices:
    """Returns corresponding DNASpec."""
    return geno.Choices(
        num_choices=self.num_choices,
        candidates=[ct.dna_spec() for ct in self._candidate_templates],
        literal_values=[self._literal_value(c)
                        for i, c in enumerate(self.candidates)],
        distinct=self.choices_distinct,
        sorted=self.choices_sorted,
        hints=self.hints,
        name=self.name,
        location=location or object_utils.KeyPath())

  def _literal_value(
      self, candidate: Any, max_len: int = 120) -> Union[int, float, str]:
    """Returns literal value for candidate."""
    if isinstance(candidate, numbers.Number):
      return candidate

    literal = object_utils.format(candidate, compact=True,
                                  hide_default_values=True,
                                  hide_missing_values=True,
                                  strip_object_id=True)
    if len(literal) > max_len:
      literal = literal[:max_len - 3] + '...'
    return literal

  def _decode(self) -> List[Any]:
    """Decode a DNA into a list of object."""
    dna = self._dna
    if self.num_choices == 1:
      # Single choice.
      if not isinstance(dna.value, int):
        raise ValueError(
            object_utils.message_on_path(
                f'Did you forget to specify values for conditional choices?\n'
                f'Expect integer for {self.__class__.__name__}. '
                f'Encountered: {dna!r}.', self.sym_path))
      if dna.value >= len(self.candidates):
        raise ValueError(
            object_utils.message_on_path(
                f'Choice out of range. Value: {dna.value!r}, '
                f'Candidates: {len(self.candidates)}.', self.sym_path))
      choices = [self._candidate_templates[dna.value].decode(
          geno.DNA(None, dna.children))]
    else:
      # Multi choices.
      if len(dna.children) != self.num_choices:
        raise ValueError(
            object_utils.message_on_path(
                f'Number of DNA child values does not match the number of '
                f'choices. Child values: {dna.children!r}, '
                f'Choices: {self.num_choices}.', self.sym_path))
      if self.choices_distinct or self.choices_sorted:
        sub_dna_values = [s.value for s in dna]
        if (self.choices_distinct
            and len(set(sub_dna_values)) != len(dna.children)):
          raise ValueError(
              object_utils.message_on_path(
                  f'DNA child values should be distinct. '
                  f'Encountered: {sub_dna_values}.', self.sym_path))
        if self.choices_sorted and sorted(sub_dna_values) != sub_dna_values:
          raise ValueError(
              object_utils.message_on_path(
                  f'DNA child values should be sorted. '
                  f'Encountered: {sub_dna_values}.', self.sym_path))
      choices = []
      for i, sub_dna in enumerate(dna):
        if not isinstance(sub_dna.value, int):
          raise ValueError(
              object_utils.message_on_path(
                  f'Choice value should be int. '
                  f'Encountered: {sub_dna.value}.',
                  object_utils.KeyPath(i, self.sym_path)))
        if sub_dna.value >= len(self.candidates):
          raise ValueError(
              object_utils.message_on_path(
                  f'Choice out of range. Value: {sub_dna.value}, '
                  f'Candidates: {len(self.candidates)}.',
                  object_utils.KeyPath(i, self.sym_path)))
        choices.append(self._candidate_templates[sub_dna.value].decode(
            geno.DNA(None, sub_dna.children)))
    return choices

  def encode(self, value: List[Any]) -> geno.DNA:
    """Encode a list of values into DNA.

    Example::

        # DNA of an object containing a single OneOf.
        # {'a': 1} => DNA(0)
        {
           'a': one_of([1, 2])
        }


        # DNA of an object containing multiple OneOfs.
        # {'b': 1, 'c': bar} => DNA([0, 1])
        {
           'b': pg.oneof([1, 2]),
           'c': pg.oneof(['foo', 'bar'])
        }

        # DNA of an object containing conditional space.
        # {'a': {'b': 1} => DNA(0, 0, 0)])
        # {'a': {'b': [4, 7]} => DNA(1, [(0, 1), 2])
        # {'a': {'b': 'bar'} => DNA(2)
        {
           'a': {
              'b': pg.oneof([
                pg.oneof([
                  pg.oneof([1, 2]),
                  pg.oneof(3, 4)]),
                  pg.manyof(2, [
                    pg.oneof([4, 5]),
                    6,
                    7
                  ]),
                ]),
                'bar',
              ])
           }
        }

    Args:
      value: A list of value that can match choice candidates.

    Returns:
      Encoded DNA.

    Raises:
      ValueError if value cannot be encoded.
    """
    if not isinstance(value, list):
      raise ValueError(
          object_utils.message_on_path(
              f'Cannot encode value: value should be a list type. '
              f'Encountered: {value!r}.', self.sym_path))
    choices = []
    if self.num_choices is not None and len(value) != self.num_choices:
      raise ValueError(
          object_utils.message_on_path(
              f'Length of input list is different from the number of choices '
              f'({self.num_choices}). Encountered: {value}.', self.sym_path))
    for v in value:
      choice_id = None
      child_dna = None
      for i, b in enumerate(self._candidate_templates):
        succeeded, child_dna = b.try_encode(v)
        if succeeded:
          choice_id = i
          break
      if child_dna is None:
        raise ValueError(
            object_utils.message_on_path(
                f'Cannot encode value: no candidates matches with '
                f'the value. Value: {v!r}, Candidates: {self.candidates}.',
                self.sym_path))
      choices.append(geno.DNA(choice_id, [child_dna]))
    return geno.DNA(None, choices)


@symbolic.members(
    [],
    init_arg_list=[
        'num_choices', 'candidates', 'choices_distinct',
        'choices_sorted', 'hints'
    ],
    # TODO(daiyip): Change to 'ManyOf' once existing code migrates to ManyOf.
    serialization_key='hyper.ManyOf',
    additional_keys=['pyglove.generators.genetic.ChoiceList']
)
class ManyOf(Choices):
  """N Choose K.

  Example::

    # Chooses 2 distinct candidates.
    v = pg.manyof(2, [1, 2, 3])

    # Chooses 2 non-distinct candidates.
    v = pg.manyof(2, [1, 2, 3], distinct=False)

    # Chooses 2 distinct candidates sorted by their indices.
    v = pg.manyof(2, [1, 2, 3], sorted=True)

    # Permutates the candidates.
    v = pg.permutate([1, 2, 3])

    # A hierarchical categorical choice:
    v2 = pg.manyof(2, [
        'foo',
        'bar',
        pg.oneof([1, 2, 3])
    ])

  See also:

    * :func:`pyglove.manyof`
    * :func:`pyglove.permutate`
    * :class:`pyglove.hyper.Choices`
    * :class:`pyglove.hyper.OneOf`
    * :class:`pyglove.hyper.Float`
    * :class:`pyglove.hyper.CustomHyper`
  """

  def custom_apply(
      self,
      path: object_utils.KeyPath,
      value_spec: pg_typing.ValueSpec,
      allow_partial: bool,
      child_transform: Optional[Callable[
          [object_utils.KeyPath, pg_typing.Field, Any], Any]] = None
      ) -> Tuple[bool, 'Choices']:
    """Validate candidates during value_spec binding time."""
    # Check if value_spec directly accepts `self`.
    if value_spec.value_type and isinstance(self, value_spec.value_type):
      return (False, self)

    if self._value_spec:
      src_spec = self._value_spec
      dest_spec = value_spec
      if not dest_spec.is_compatible(src_spec):
        raise TypeError(
            object_utils.message_on_path(
                f'Cannot bind an incompatible value spec {dest_spec!r} '
                f'to {self.__class__.__name__} with bound spec {src_spec!r}.',
                path))
      return (False, self)

    list_spec = typing.cast(
        pg_typing.List,
        pg_typing.ensure_value_spec(
            value_spec, pg_typing.List(pg_typing.Any()), path))
    if list_spec:
      for i, c in enumerate(self.candidates):
        list_spec.element.value.apply(
            c,
            self._allow_partial,
            root_path=path + f'candidates[{i}]')
    self._value_spec = list_spec
    return (False, self)


@symbolic.members(
    [
        ('num_choices', 1)
    ],
    init_arg_list=['candidates', 'hints', 'where'],
    serialization_key='hyper.OneOf',
    additional_keys=['pyglove.generators.genetic.ChoiceValue']
)
class OneOf(Choices):
  """N Choose 1.

  Example::

    # A single categorical choice:
    v = pg.oneof([1, 2, 3])

    # A hierarchical categorical choice:
    v2 = pg.oneof([
        'foo',
        'bar',
        pg.oneof([1, 2, 3])
    ])

  See also:

    * :func:`pyglove.oneof`
    * :class:`pyglove.hyper.Choices`
    * :class:`pyglove.hyper.ManyOf`
    * :class:`pyglove.hyper.Float`
    * :class:`pyglove.hyper.CustomHyper`
  """

  def _on_bound(self):
    """Event triggered when members are bound."""
    super()._on_bound()
    assert self.num_choices == 1

  def _decode(self) -> Any:
    """Decode a DNA into an object."""
    return super()._decode()[0]

  def encode(self, value: Any) -> geno.DNA:
    """Encode a value into a DNA."""
    # NOTE(daiyip): Single choice DNA will automatically be pulled
    # up from children to current node. Thus we simply returns
    # encoded DNA from parent node.
    return super().encode([value])

  def custom_apply(
      self,
      path: object_utils.KeyPath,
      value_spec: pg_typing.ValueSpec,
      allow_partial: bool,
      child_transform: Optional[Callable[
          [object_utils.KeyPath, pg_typing.Field, Any], Any]] = None
      ) -> Tuple[bool, 'OneOf']:
    """Validate candidates during value_spec binding time."""
    # Check if value_spec directly accepts `self`.
    if value_spec.value_type and isinstance(self, value_spec.value_type):
      return (False, self)

    if self._value_spec:
      if not value_spec.is_compatible(self._value_spec):
        raise TypeError(
            object_utils.message_on_path(
                f'Cannot bind an incompatible value spec {value_spec!r} '
                f'to {self.__class__.__name__} with bound '
                f'spec {self._value_spec!r}.', path))
      return (False, self)

    for i, c in enumerate(self.candidates):
      value_spec.apply(
          c,
          self._allow_partial,
          root_path=path + f'candidates[{i}]')
    self._value_spec = value_spec
    return (False, self)

#
# Helper methods for creating hyper values.
#


def oneof(candidates: Iterable[Any],
          *,
          name: Optional[str] = None,
          hints: Optional[Any] = None) -> Any:
  """N choose 1.

  Example::

    @pg.members([
      ('x', pg.typing.Int())
    ])
    class A(pg.Object):
      pass

    # A single categorical choice:
    v = pg.oneof([1, 2, 3])

    # A complex type as candidate.
    v1 = pg.oneof(['a', {'x': 1}, A(1)])

    # A hierarchical categorical choice:
    v2 = pg.oneof([
        'foo',
        'bar',
        A(pg.oneof([1, 2, 3]))
    ])

  See also:

    * :class:`pyglove.hyper.OneOf`
    * :func:`pyglove.manyof`
    * :func:`pyglove.floatv`
    * :func:`pyglove.permutate`
    * :func:`pyglove.evolve`

  .. note::

    Under symbolic mode (by default), `pg.oneof` returns a ``pg.hyper.OneOf``
    object. Under dynamic evaluation mode, which is called under the context of
    :meth:`pyglove.hyper.DynamicEvaluationContext.collect` or
    :meth:`pyglove.hyper.DynamicEvaluationContext.apply`, it evaluates to
    a concrete candidate value.

    To use conditional search space in dynamic evaluation mode, the candidate
    should be wrapped with a `lambda` function, which is not necessary under
    symbolic mode. For example::

      pg.oneof([lambda: pg.oneof([0, 1], name='sub'), 2], name='root')

  Args:
    candidates: Candidates to select from. Items of candidate can be any type,
      therefore it can have nested hyper primitives, which forms a hierarchical
      search space.
    name: A name that can be used to identify a decision point in the search
      space. This is needed when the code to instantiate the same hyper
      primitive may be called multiple times under a
      `pg.DynamicEvaluationContext.collect` context or under a
      `pg.DynamicEvaluationContext.apply` context.
    hints: An optional value which acts as a hint for the controller.

  Returns:
    In symbolic mode, this function returns a `ChoiceValue`.
    In dynamic evaluation mode, this function returns one of the items in
    `candidates`.
    If evaluated under a `pg.DynamicEvaluationContext.apply` scope,
    this function will return the selected candidate.
    If evaluated under a `pg.DynamicEvaluationContext.collect`
    scope, it will return the first candidate.
  """
  return OneOf(candidates=list(candidates), name=name, hints=hints)


def manyof(k: int,
           candidates: Iterable[Any],
           distinct: bool = True,
           sorted: bool = False,    # pylint: disable=redefined-builtin
           *,
           name: Optional[str] = None,
           hints: Optional[Any] = None,
           **kwargs) -> Any:
  """N choose K.

  Example::

    @pg.members([
      ('x', pg.typing.Int())
    ])
    class A(pg.Object):
      pass

    # Chooses 2 distinct candidates.
    v = pg.manyof(2, [1, 2, 3])

    # Chooses 2 non-distinct candidates.
    v = pg.manyof(2, [1, 2, 3], distinct=False)

    # Chooses 2 distinct candidates sorted by their indices.
    v = pg.manyof(2, [1, 2, 3], sorted=True)

    # A complex type as candidate.
    v1 = pg.manyof(2, ['a', {'x': 1}, A(1)])

    # A hierarchical categorical choice:
    v2 = pg.manyof(2, [
        'foo',
        'bar',
        A(pg.oneof([1, 2, 3]))
    ])

  .. note::

    Under symbolic mode (by default), `pg.manyof` returns a ``pg.hyper.ManyOf``
    object. Under dynamic evaluation mode, which is called under the context of
    :meth:`pyglove.hyper.DynamicEvaluationContext.collect` or
    :meth:`pyglove.hyper.DynamicEvaluationContext.apply`, it evaluates to
    a concrete candidate value.

    To use conditional search space in dynamic evaluate mode, the candidate
    should be wrapped with a `lambda` function, which is not necessary under
    symbolic mode. For example::

        pg.manyof(2, [
           lambda: pg.oneof([0, 1], name='sub_a'),
           lambda: pg.floatv(0.0, 1.0, name='sub_b'),
           lambda: pg.manyof(2, ['a', 'b', 'c'], name='sub_c')
        ], name='root')

  See also:

    * :class:`pyglove.hyper.ManyOf`
    * :func:`pyglove.manyof`
    * :func:`pyglove.floatv`
    * :func:`pyglove.permutate`
    * :func:`pyglove.evolve`

  Args:
    k: number of choices to make. Should be no larger than the length of
      `candidates` unless `choice_distinct` is set to False,
    candidates: Candidates to select from. Items of candidate can be any type,
      therefore it can have nested hyper primitives, which forms a hierarchical
      search space.
    distinct: If True, each choice needs to be unique.
    sorted: If True, choices are sorted by their indices in the
      candidates.
    name: A name that can be used to identify a decision point in the search
      space. This is needed when the code to instantiate the same hyper
      primitive may be called multiple times under a
      `pg.DynamicEvaluationContext.collect` context or a
      `pg.DynamicEvaluationContext.apply` context.
    hints: An optional value which acts as a hint for the controller.
    **kwargs: Keyword arguments for backward compatibility.
      `choices_distinct`: Old name for `distinct`.
      `choices_sorted`: Old name for `sorted`.

  Returns:
    In symbolic mode, this function returns a `Choices`.
    In dynamic evaluate mode, this function returns a list of items in
    `candidates`.
    If evaluated under a `pg.DynamicEvaluationContext.apply` scope,
    this function will return a list of selected candidates.
    If evaluated under a `pg.DynamicEvaluationContext.collect`
    scope, it will return a list of the first valid combination from the
    `candidates`. For example::

        # Evaluates to [0, 1, 2].
        manyof(3, range(5))

        # Evaluates to [0, 0, 0].
        manyof(3, range(5), distinct=False)
  """
  choices_distinct = kwargs.pop('choices_distinct', distinct)
  choices_sorted = kwargs.pop('choices_sorted', sorted)
  return ManyOf(
      num_choices=k,
      candidates=list(candidates),
      choices_distinct=choices_distinct,
      choices_sorted=choices_sorted,
      name=name,
      hints=hints)


def permutate(candidates: Iterable[Any],
              name: Optional[str] = None,
              hints: Optional[Any] = None) -> Any:
  """Permuatation of candidates.

  Example::

    @pg.members([
      ('x', pg.typing.Int())
    ])
    class A(pg.Object):
      pass

    # Permutates the candidates.
    v = pg.permutate([1, 2, 3])

    # A complex type as candidate.
    v1 = pg.permutate(['a', {'x': 1}, A(1)])

    # A hierarchical categorical choice:
    v2 = pg.permutate([
        'foo',
        'bar',
        A(pg.oneof([1, 2, 3]))
    ])

  .. note::

    Under symbolic mode (by default), `pg.manyof` returns a ``pg.hyper.ManyOf``
    object. Under dynamic evaluate mode, which is called under the context of
    :meth:`pyglove.hyper.DynamicEvaluationContext.collect` or
    :meth:`pyglove.hyper.DynamicEvaluationContext.apply`, it evaluates to
    a concrete candidate value.

    To use conditional search space in dynamic evaluate mode, the candidate
    should be wrapped with a `lambda` function, which is not necessary under
    symbolic mode. For example::

      pg.permutate([
         lambda: pg.oneof([0, 1], name='sub_a'),
         lambda: pg.floatv(0.0, 1.0, name='sub_b'),
         lambda: pg.manyof(2, ['a', 'b', 'c'], name='sub_c')
      ], name='root')

  See also:

    * :class:`pyglove.hyper.ManyOf`
    * :func:`pyglove.oneof`
    * :func:`pyglove.manyof`
    * :func:`pyglove.floatv`
    * :func:`pyglove.evolve`

  Args:
    candidates: Candidates to select from. Items of candidate can be any type,
      therefore it can have nested hyper primitives, which forms a hierarchical
      search space.
    name: A name that can be used to identify a decision point in the search
      space. This is needed when the code to instantiate the same hyper
      primitive may be called multiple times under a
      `pg.DynamicEvaluationContext.collect` context or a
      `pg.DynamicEvaluationContext.apply` context.
    hints: An optional value which acts as a hint for the controller.

  Returns:
    In symbolic mode, this function returns a `Choices`.
    In dynamic evaluate mode, this function returns a permutation from
    `candidates`.
    If evaluated under an `pg.DynamicEvaluationContext.apply` scope,
    this function will return a permutation of candidates based on controller
    decisions.
    If evaluated under a `pg.DynamicEvaluationContext.collect`
    scope, it will return the first valid permutation.
    For example::

      # Evaluates to [0, 1, 2, 3, 4].
      permutate(range(5), name='numbers')
  """
  candidates = list(candidates)
  return manyof(
      len(candidates), candidates,
      choices_distinct=True, choices_sorted=False, name=name, hints=hints)
